/*
// THIS FILE IS COPYRIGHT 1994 BY TELENOR RESEARCH, NORWAY
//
//                    ALL RIGHTS RESERVED.
//
// For rights granted to the public see the COPYRIGHT file
*/

/* Programming : Jon Stephenson von Tetzchner		 */

#include "trans.h"

int in_table = FALSE;
int in_table_title = FALSE;
int in_table_header = FALSE;
int in_table_body = FALSE;
int in_table_footer = FALSE;

struct FMTable *first_table = NULL;
struct FMTable *current_table = NULL;
struct FMRow *current_row = NULL;
struct FMCell *current_cell = NULL;
struct FMCellLine *current_cell_line = NULL;

int column = 0;		/* current column */
int gap = 2;		/* number of spaces separating columns */

FILE *tablef = NULL;
char table_filename[300];
int table_count = 0;
char *table_filenames[1000];


/******************************************************************************
  void GetTableID() :
 *****************************************************************************/

void GetTableID()
{
    current_table->id = GetID();

    if (NOGRAPHICS)
	return;

    sprintf(table_filename, "%s.tbl_%d.mif", TRANSFilename, current_table->id);
    sprintf(sys_call, "cp %s %s", HEADFilename, table_filename);
    if (!QUIET)
        fprintf(stderr, "trans: Making FrameMaker MIF file \"%s\"\n",
		table_filename);
    system(sys_call);

    tablef = fopen(table_filename, "a");
    if (!tablef)
	{
	    fprintf(stderr, "trans: ERROR: Could not open file \"%s\"\n",
		    table_filename);
	    exit(-1);
	}

    table_filenames[table_count++] = (char *) strdup(table_filename);

    fprintf(tablef, "<Tbls \n");
    fprintf(tablef, " <Tbl \n");
    fprintf(tablef, " <TblID %d>\n", current_table->id);
}


/******************************************************************************
  void GetTableNumColumns() :
 *****************************************************************************/

void GetTableNumColumns()
{
    int i;

    current_table->num_columns = GetID();
    current_table->column_widths =
        (int *) calloc(current_table->num_columns, sizeof(int));
    for (i = 0; i < current_table->num_columns; i++)
        current_table->column_widths[i] = 0;
}


/******************************************************************************
  void TableStart() :
 *****************************************************************************/

void TableStart()
{
    debug("TableStart\n");

    in_table = TRUE;
    current_table = (struct FMTable *) calloc(1, sizeof(struct FMTable));
    current_table->header_end[0] = 0;
    current_table->body_end[0] = 0;
    current_table->footer_end[0] = 0;
}


/******************************************************************************
  void TableEnd() :
 *****************************************************************************/

void TableEnd()
{
    int i;

    debug("TableEnd\n");

    in_table = FALSE;

    for (i = 1; i <= footnote_def_count; i++)
        current_table->footnote[i] = footnote[i];

    current_table->footnote_def_count = footnote_def_count;
    footnote_def_count = 0;
    footnote_use_count = 0;
    current_table->next_table = first_table;
    first_table = current_table;
    current_table = NULL;

    if (NOGRAPHICS)
	return;

    fprintf(tablef, "> # end of Tbl\n");
    fprintf(tablef, "> # end of Tbls\n");

    fprintf(tablef, "<TextFlow \n");
    fprintf(tablef, " <Para \n");
    fprintf(tablef, "  <PgfTag `fm2htmlparagraphformat'>\n");
    fprintf(tablef, "  <ParaLine \n");
    fprintf(tablef, "   <ATbl %d>\n", first_table->id);
    fprintf(tablef, "  >\n");
    fprintf(tablef, " > # end of Para\n");
    fprintf(tablef, "> # end of TextFlow\n");
    fprintf(tablef, "# End of MIFFile\n");

    fclose(tablef);
    tablef = NULL;
}


/******************************************************************************
  void TableTitleStart() :
 *****************************************************************************/

void TableTitleStart()
{
    debug("TableTitleStart\n");

    in_table_title = TRUE;
}


/******************************************************************************
  void TableTitleEnd() :
 *****************************************************************************/

void TableTitleEnd()
{
    debug("TableTitleEnd\n");

    in_table_title = FALSE;
}


/******************************************************************************
  void TableHeadStart() :
 *****************************************************************************/

void TableHeadStart()
{
    debug("TableHeadStart\n");

    in_table_header = TRUE;
}


/******************************************************************************
  void TableHeadEnd() :
 *****************************************************************************/

void TableHeadEnd()
{
    debug("TableHeadEnd\n");

    in_table_header = FALSE;

    if (font_bold)
        {
            strcat(current_table->header_end, "<TableHeadEnd></B>");
	    font_bold = FALSE;
	}
    if (font_italic)
        {
	    strcat(current_table->header_end, "</I>");
	    font_italic = FALSE;
	}
    if (font_underline)
        {
	    strcat(current_table->header_end, "</U>");
	    font_underline = FALSE;
	}
    if (font_subscript)
        {
	    strcat(current_table->header_end, "</EM>");
	    font_subscript = FALSE;
	}
    if (font_superscript)
        {
	    strcat(current_table->header_end, "</EM>");
	    font_superscript = FALSE;
	}
    if (in_code)
        {
	    strcat(current_table->header_end, "</CODE>");
	    in_code = FALSE;
	}
}


/******************************************************************************
  void TableBodyStart() :
 *****************************************************************************/

void TableBodyStart()
{
    debug("TableBodyStart\n");

    in_table_body = TRUE;
}


/******************************************************************************
  void TableBodyEnd() :
 *****************************************************************************/

void TableBodyEnd()
{
    debug("TableBodyEnd\n");

    in_table_body = FALSE;
}


/******************************************************************************
  void TableFootStart() :
 *****************************************************************************/

void TableFootStart()
{
    debug("TableFootStart\n");

    in_table_footer = TRUE;
}


/******************************************************************************
  void TableFootEnd() :
 *****************************************************************************/

void TableFootEnd()
{
    debug("TableFootEnd\n");

    in_table_footer = FALSE;

    if (font_bold)
        {
            strcat(current_table->footer_end, "<TableFootEnd></B>");
	    font_bold = FALSE;
	}
    if (font_italic)
        {
	    strcat(current_table->footer_end, "</I>");
	    font_italic = FALSE;
	}
    if (font_underline)
        {
	    strcat(current_table->footer_end, "</U>");
	    font_underline = FALSE;
	}
    if (font_subscript)
        {
	    strcat(current_table->footer_end, "</EM>");
	    font_subscript = FALSE;
	}
    if (font_superscript)
        {
	    strcat(current_table->footer_end, "</EM>");
	    font_superscript = FALSE;
	}
    if (in_code)
        {
	    strcat(current_table->footer_end, "</CODE>");
	    in_code = FALSE;
	}
}


/******************************************************************************
  void TableRowStart() :
 *****************************************************************************/

void TableRowStart()
{
    struct FMRow *new_row = (struct FMRow *) calloc(1, sizeof(struct FMRow));
    struct FMRow *row;

    if (in_table_header)
        if (current_table->header)
	    current_row->next_row = new_row;
        else
	    current_table->header = new_row;
    else if (in_table_footer)
        if (current_table->footer)
	    current_row->next_row = new_row;
        else
	    current_table->footer = new_row;
    else
        if (current_table->row)
	    current_row->next_row = new_row;
        else
	    current_table->row = new_row;

    current_row = new_row;
}    


/******************************************************************************
  void TableRowEnd() :
 *****************************************************************************/

void TableRowEnd()
{
    column = 0;
}


/******************************************************************************
  void TableCellStart() :
 *****************************************************************************/

void TableCellStart()
{
    struct FMCell *new_cell = (struct FMCell *) calloc(1, sizeof(struct FMCell));
    if (!current_row->cell)
	current_cell = current_row->cell = new_cell;
    else
	{
	    current_cell->next_cell = new_cell;
	    current_cell = new_cell;
	}	
    current_cell->cell_columns = 1;
    current_cell->alignment = 'L';
}	


/******************************************************************************
  void TableCellColumns() :
 *****************************************************************************/

void TableCellColumns()
{
    current_cell->cell_columns = GetID();
}	


/******************************************************************************
  void TableCellEnd() :
 *****************************************************************************/

void TableCellEnd()
{
    ParaEnd();
    column++;
}


/******************************************************************************
  int GetID() :
        Input  : Uses yytext
	Output : Number not quoted.
 *****************************************************************************/

int GetID()
{
    int i = 0;

    while(yytext[i] && yytext[i] != 32)
        i++;

    if (yytext[i] == 32)
	{
	    i++;
	    return atoi(&(yytext[i]));
	}			
    else
        return 0;
}


/******************************************************************************
 void StringLength() : Returns length of string, conting special characters
                       like "\xbf " one. (and html tags like 0...)
 *****************************************************************************/

int StringLength(str)
      char* str;
{
    int i = 0;
    int lth = 0;

    if (str)
	while (str[i]) {
	    if (str[i] == '\\')
	        i += 4;
	    else if (str[i] == '<') {
		int j = i++;
		while (str[i] && str[i] != '>')
		    i++;
		if (!str[i])
		    i = j;
		else
		    lth--;
	    }
	    else if (str[i] == '&') {
		if ((str[i+1] == 'g' || str[i+1] == 'l') && str[i+2] == 't' &&
		    str[i+3] == ';')
		    i += 3;
		else if (str[i+1] == 'a' && str[i+2] == 'm' && 
			 str[i+3] == 'p' && str[i+4] == ';')
		    i += 4;
		else
		    fprintf(stderr,
			    "trans: ERROR: Unknown HTML escape string: \"%s\"\n",
			    str[i]);
	        }
	    i++;
	    lth++;
	}

    return lth;
}


/******************************************************************************
 void NewCellLine() : 
 *****************************************************************************/

void NewCellLine()
{
    if (in_string)
        PrintHeaderTypeEnd();

    if (current_cell)
	{
	    struct FMCellLine *new_line =
		(struct FMCellLine *) calloc(1, sizeof(struct FMCellLine));
	    current_cell->num_lines++;

	    if (current_cell->num_lines > current_row->max_num_lines)
		current_row->max_num_lines = current_cell->num_lines;

	    new_line->fmline = NULL;

	    if (!current_cell->line) 
		current_cell->line = new_line;
	    else
		{
		    struct FMCellLine *temp1, *temp2;
		    temp1 = current_cell->line;
		    temp2 = temp1->next_line;
		    while(temp2)
			{
			    temp1 = temp2;
			    temp2 = temp1->next_line;
			}
		    temp1->next_line = new_line;
		}
	    current_cell_line = new_line;
	}
}


/******************************************************************************
 void UseString() : 
 *****************************************************************************/

void UseString(string_contents)
      char *string_contents;
{
    int str_lth;

    if (in_footnote)
	{
	    if (footnote_def_count >= MAXNUMFOOTNOTES)
		{
		    fprintf(stderr, "\nERROR: Too many footnotes.\n\n");
		    return;
		}
	    strcat(footnote[footnote_def_count], string_contents);
	}		    
    else if (in_table_title)
	{
	    if (current_table->table_title)
		{
		    char *temp1;
		    temp1 = current_table->table_title;
		    current_table->table_title = calloc(strlen(temp1) + 
							strlen(string_contents) + 1, sizeof(char));
		    strcpy(current_table->table_title, temp1);
		    strcat(current_table->table_title, string_contents);
		    free(temp1);
		}
	    else
		{
		    current_table->table_title = calloc(strlen(para_num_string) + 
							strlen(string_contents) + 1, sizeof(char));
		    strcat(current_table->table_title, string_contents);
		}
	}
    else if (current_cell && current_cell_line)
	{
	    if (current_cell_line->fmline)
		{
		    char *temp1;
		    temp1 = current_cell_line->fmline;
		    current_cell_line->fmline = 
		        calloc(strlen(temp1) + strlen(string_contents) + 1,
			       sizeof(char));
		    strcpy(current_cell_line->fmline, temp1);
		    strcat(current_cell_line->fmline, string_contents);
		    free(temp1);
		}
	    else
		{
		    current_cell_line->fmline =
		        calloc(strlen(string_contents) + 1, sizeof(char));
		    strcpy(current_cell_line->fmline, string_contents);
		}

	    str_lth = StringLength(current_cell_line->fmline) + gap;

	    if (current_cell->cell_columns == 1)
	        {
		    if (str_lth > current_table->column_widths[column])
		        current_table->column_widths[column] = str_lth;
		}
	    else
	        {
		    int i, width;
		    
		    for (i = 0, width = 0; i < current_cell->cell_columns; i++)
		        width += current_table->column_widths[column+i];

		    if (str_lth > width)
		        {
			    int add = (str_lth - width) /
			        current_cell->cell_columns + 1;

			    for (i = 0; i < current_cell->cell_columns; i++)
			        current_table->column_widths[column+i] += add;
			}
		}
	}
}


/******************************************************************************
 void PrintTableTitle() : 
 *****************************************************************************/

void PrintTableTitle(f, table, max_lth)
      FILE *f;
      struct FMTable *table;
      int max_lth;
{
    int i, indent;
    char *str = table->table_title;
    char line[500];
    char tmpline[500];
    char *word;
    int IsNotFirstLine = 0;

    debug("PrintTableTitle\n");

    tmpline[0] = 0;
    line[0] = 0;

    word = strtok(str, " \t");
    while (word)
	{
	    strcat(tmpline, word);
	    strcat(tmpline, " ");
	    if (StringLength(tmpline) < max_lth)
		{
		    strcat(line, word);
		    strcat(line, " ");
		}
	    else
		{
		    if (strlen(line) > 0)
			{
			    if (IsNotFirstLine) fprintf(f, "\t\t");
			    fprintf(f, "%s\n", line);
			    strcpy(line, word);
			    tmpline[0] = 0;
			    IsNotFirstLine = 1;
			}
		    else
			{
			    if (IsNotFirstLine) fprintf(f, "\t\t");
			    fprintf(f, "%s\n", tmpline);
			    tmpline[0] = 0;
			    IsNotFirstLine = 1;
			}
		}
	    word = strtok(NULL, " \t");
	}

    if (strlen(line) > 0)
	{
	    if (IsNotFirstLine)
	        fprintf(f, "\t\t");
	    fprintf(f, "%s\n", line);
	}
}


/******************************************************************************
 void InsertTable() : 
 *****************************************************************************/

void InsertTable(f, id)
      FILE *f;
      int id;
{
    struct FMTable *table;
    struct FMRow *row;
    int table_width = 0;
    int i;

    debug("InsertTable\n");

    table = FindTable(id);	/* Find the table to print */

    if (!table)
	{
	    fprintf(stderr, "trans: ERROR: Unknown table ID: %d\n", id);
	    exit(-1);
	}

    for (i = 0; i < table->num_columns; i++)
	table_width += table->column_widths[i];

    if (table->table_title)       /* Print the table title */
    {
        /* look out, here comes a kludge! (i want the title to be selectable,
	   not just the icon) */
        if (strncmp(table->table_title, "<A NAME=", strlen("A NAME=")) == 0)
	{
	    char *orig = table->table_title;
	    char beginning[1000];
	    char *remainder = table->table_title;
	    char *name = &table->table_title[3];

	    while (*remainder != '>')
	        remainder++;
	    *remainder = 0;	/* terminates name */
	    remainder++;

	    sprintf(beginning,
		"<P><A HREF=\"%s.tbl_%d.ps\" %s><IMG SRC=\"fm2html-table.gif\">",
		TRANSBasename, table->id, name);

	    table->table_title = (char *) calloc(strlen(beginning) +
						 strlen(remainder),
						 sizeof(char));
	    sprintf(table->table_title, "%s%s", beginning, remainder);

	    free(orig);
	    PrintTableTitle(f, table, 80);
	}
	else
	{
	    fprintf(f, "<P><A HREF=\"%s.tbl_%d.ps\"><IMG SRC=\"fm2html-table.gif\">",
		    TRANSBasename, table->id);
	    PrintTableTitle(f, table, 80);
	    fprintf(f, "</A>\n");
	}
    }
    else
        fprintf(f, "<P><A HREF=\"%s.tbl_%d.ps\"><IMG SRC=\"fm2html-table.gif\"></A>",
		TRANSBasename, table->id);

    fprintf(f, "<PRE>\n");

    for (i = 0; i < table_width - 2; i++)	/* Print "-------------" */
	fputc('-', f);
    fputc('\n', f);

    row = table->header;
    while (row)					/* Print the header rows */
	{
	    PrintTableRow(f, row, table->column_widths);
	    row = row->next_row;
	}

    if (table->header)
        fprintf(f, "%s", table->header_end);

    if (table->header)				/* Print "--------------" */
	{
	    for (i = 0; i < table_width - 2; i++)
		fputc('-',f );
	    fputc('\n', f);
	}

    row = table->row;
    while (row)					/* Print the table rows */
	{
	    PrintTableRow(f, row, table->column_widths);
	    row = row->next_row;
	}

    for (i = 0; i < table_width - 2; i++)	/* Print "--------------" */
	fputc('-', f);
    fputc('\n', f);

    row = table->footer;
    while (row)					/* Print the footer rows */
	{
	    PrintTableRow(f, row, table->column_widths);
	    row = row->next_row;
	}

    if (table->footer)
        fprintf(f, "%s", table->footer_end);

    if (table->footer)				/* Print "--------------" */
	{
	    for (i = 0; i < table_width - 2; i++)
		fputc('-',f );
	    fputc('\n', f);
	}

    fprintf(f, "</PRE>\n");

    if (table->footnote_def_count)
	PrintTableFootnotes(f, table);
    table->footnote_def_count = 0;
}
	

/******************************************************************************
 void PrintTableRow() : 
 *****************************************************************************/
    
void PrintTableRow(f, row, column_widths)
      FILE *f;
      struct FMRow *row;
      int *column_widths;
{
    int line;
    
    for (line = 0; line < row->max_num_lines; line++)
	{
	    struct FMCell *fmcell = row->cell;
	    int col = 0;

	    while (fmcell)
		{
		    int i, width;
		    int cell_columns = fmcell->cell_columns;

		    for (i = 0, width = 0; i < cell_columns; i++)
		        width += column_widths[col+i];

		    PrintTableCell(f, fmcell, line, width);

		    /* Skip past straddled columns to next actual column */
		    for (i = 0; i < cell_columns; i++)
		        fmcell = fmcell->next_cell, col++;
		}
	    fprintf(f, "\n");
	}
}


/******************************************************************************
 void PrintTableCell() : 
 *****************************************************************************/
    
void PrintTableCell(f, cell, line_number, width)
      FILE *f;
      struct FMCell *cell;
      int line_number;
      int width;
{
    struct FMCellLine *line = cell->line;
    char formatstring[20];
    int i = 0;

    while (line && i++ < line_number)
        line = line->next_line;

    if (line && line->fmline)
	{
            width -= gap;

	    /* Add room for HTML tags and special characters */
	    width += (strlen(line->fmline) - StringLength(line->fmline));

	    if (cell->alignment == 'R')
	        {
		    sprintf(formatstring, "%%%ds", width);
		    fprintf(f, formatstring, line->fmline);
	        }
	    else if (cell->alignment == 'C')
	        {
		    int indent = (width - strlen(line->fmline)) / 2;
		    sprintf(formatstring, "%%%ds%%%ds", width - indent, 
			    indent);
		    fprintf(f, formatstring, line->fmline, " ");
	        }
	    else
	        {
		    sprintf(formatstring, "%%-%ds", width);
		    fprintf(f, formatstring, line->fmline);
		}
	    sprintf(formatstring, "%%%ds", gap);
	    fprintf(f, formatstring, " ");
	}
    else
	{
	    sprintf(formatstring, "%%%ds", width);
	    fprintf(f, formatstring, " ");
	}
}


/******************************************************************************
 FMTable *FindTable() : 
 *****************************************************************************/

struct FMTable *FindTable(id)
      int id;
{
    struct FMTable *stable = first_table;

    while (stable && stable->id != id)
	stable = stable->next_table;

    if (stable)
        return stable;
    return NULL;
}


/******************************************************************************
 void PrintTableFootnotes() : 
 *****************************************************************************/

void PrintTableFootnotes(f, table)
      FILE *f;
      struct FMTable *table;
{
    int i;
    fprintf(f, "<DL COMPACT>\n");
    for (i = 1; i <= table->footnote_def_count; i++)
	{
	    fprintf(f, "<DT><A NAME=FN%d%c>(%c)</A><DD>%s\n", table->id,  
		    i + 'a' - 1, i + 'a' - 1, table->footnote[i]);
	    free(table->footnote[i]);
	}
    fprintf(f, "</DL>\n");
}
